#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   utils.py
@Time    :   2021/04/16 12:29:39
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import os
import sys
import pickle
import time

import ruamel.yaml
import pandas as pd
from scipy import io


def add_time(func):
    def wrapper(*args, **kw):
        time_now = time.strftime("%Y-%m-%d %H:%M:%S")
        sys.stdout.write(f'INFO {time_now} ')
        return func(*args, **kw)
    return wrapper

@add_time
def print_log(string, end="\n"):
    print(string, end=end)

def excute_info(start_info=None, end_info=None):
    def call(func):
        def wrapper(*args, **kwargs):
            if start_info:
                print_log(start_info)
            ret = func(*args, **kwargs)
            if end_info:
                print_log(end_info)
            return ret
        return wrapper
    return call

def read_SingleCellExperiment_rds(input_RDS):
    import anndata2ri
    from rpy2.robjects import r
    anndata2ri.activate()
    rscript = f'readRDS("{input_RDS}")'
    adata = r(rscript)
    return adata

def read_config():
    yaml = ruamel.yaml.YAML()
    CONFIG_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),'..','conf'))
    CONFIG_YML_PATH = os.path.join(CONFIG_PATH,'config.yml')
    with open(CONFIG_YML_PATH, 'r') as config_file:
        CONFIG = yaml.load(config_file.read())
    return CONFIG, CONFIG_PATH

def store_to_pickle(obj, path):
    with open(os.path.join(path), 'wb+') as f:
        pickle.dump(obj, f)


def read_pickle(path):
    with open(os.path.join(path), 'rb') as f:
        ret = pickle.load(f)
    return ret

def safe_makedirs(path):
    if not os.path.exists(path):
        os.makedirs(path)


def write_to_mtx(data, path):
    if not os.path.exists(path):
        os.makedirs(path)
    pd.DataFrame(data.var.index, data.var.index).to_csv(os.path.join(path, "genes.tsv"), sep="\t", header=False)
    pd.DataFrame(data.obs.index).to_csv(os.path.join(path, "barcodes.tsv"), sep = "\t", index=False, header=False)
    data.obs.to_csv(os.path.join(path, "metadata.tsv"), sep = "\t", index=False, header=False)
    io.mmwrite(os.path.join(path, "matrix.mtx"), data.X.T)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   post_processing.py
@Time    :   2021/04/16 12:34:26
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''


import numpy as np
import pandas as pd
# from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, wait, ALL_COMPLETED
from SCRIP.utilities.utils import excute_info



@excute_info('Generating merged anndata ... ', 'Finished Generating merged anndata!')
def merge_giggle_adata(adata, table, data_type, table2=''):
    # table : each row is a cell, each column is a TF/Gene
    new_adata = adata.copy()
    if data_type == 'integration':
        d_table = table.copy()
        d_table2 = table2.copy()
        new_adata.uns['ChIP'] = d_table
        new_adata.uns['motif'] = d_table2
        # extract factor
        chip_tfs = d_table.columns.tolist()
        motif_tfs = d_table2.columns.tolist()
        overlap_tf_list = list(set(chip_tfs).intersection(set(motif_tfs)))
        chip_other_tf_list = list(set(chip_tfs) - set(overlap_tf_list))
        motif_other_tf_list = list(set(motif_tfs) - set(overlap_tf_list))
        ovlp_chip_table = d_table[overlap_tf_list]
        ovlp_motif_table = d_table2[overlap_tf_list]
        unique_chip_table = d_table[chip_other_tf_list]
        unique_motif_table = d_table2[motif_other_tf_list]
        ovlp_final_table = pd.DataFrame()
        for ovlp_factor in overlap_tf_list:
            if np.std(ovlp_chip_table[ovlp_factor]) >= np.std(ovlp_motif_table[ovlp_factor]):
                ovlp_final_table[ovlp_factor] = ovlp_chip_table[ovlp_factor]
            else:
                ovlp_final_table[ovlp_factor] = ovlp_motif_table[ovlp_factor]

        final_table = pd.concat([ovlp_final_table,unique_chip_table,unique_motif_table], axis=1)
        final_table.columns = ['I_' + tf for tf in final_table.columns.tolist()]
        new_adata.uns['integrated'] = final_table
        new_adata.obs = pd.concat([new_adata.obs, final_table.reindex(new_adata.obs.index.tolist())], axis=1)
    else:
        d_table = table.copy()
        if data_type == 'ChIP-seq':
            new_adata.uns['ChIP'] = d_table
            d_table.columns = ['C_' + tf for tf in d_table.columns.tolist()]
        elif data_type == 'motif':
            new_adata.uns['motif'] = d_table
            d_table.columns = ['M_' + tf for tf in d_table.columns.tolist()]


        new_adata.obs = pd.concat([new_adata.obs, d_table.reindex(new_adata.obs.index.tolist())], axis=1)

    return new_adata



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   utils.py
@Time    :   2021/04/16 12:34:42
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import os
import sys
from datetime import datetime, timedelta
import json
from SCRIP.Constants import SCRIP_VERSION
from SCRIP.utilities.utils import store_to_pickle, read_pickle


def time_estimate(cell_number, core, chip_factor_number=4499):
#   generate bed: 0.5 second per cell,  search index: 10 seconds per cell, cal p: 2 seconds per factor
    # chip_process = True if reference_method in ['integration', 'both', 'chip'] else False
    # motif_process = True if reference_method in ['integration', 'both', 'motif'] else False
    seconds = int(cell_number * core / 50) # single core process, find nearest cells function
    seconds += 3 * cell_number
    # if chip_process == True:
    seconds += 10 * cell_number
    seconds += 2 * chip_factor_number
    # if motif_process == True:
    #     seconds += 10 * cell_number
    #     seconds += 2 * motif_factor_number

    seconds = seconds / core
    # now_time = datetime.now()
    future_time = (datetime.now() + timedelta(seconds=seconds)).strftime("%Y-%m-%d %H:%M:%S")
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    elapse = "{hour:0>2d}:{minute:0>2d}:{second:0>2d}".format(hour=int(h), minute=int(m), second=int(s))
    return elapse, future_time

class EnrichRunInfo(object):
    def __init__(self, file_path, parameters_dict):
        if os.path.exists(file_path):
            with open(file_path, 'r') as jf:
                info = json.load(jf)
        else:
            info = {}
            info['version'] = SCRIP_VERSION
            info['file_path'] = os.path.abspath(file_path)
            info['project_folder'] = os.path.abspath(os.path.dirname(file_path))
            info['parameters'] = parameters_dict
            info['progress'] = {

                'bed_generation': 'No',
                'bed_search': 'No',

                'dataset_overlap_df_store': 'No',
                'dataset_cell_norm_df_store': 'No',
                'dataset_score_source_df_store': 'No',
                'tf_cell_score_df_store': 'No',

                'result_store': 'No'
            }
        self.version = info['version']
        self.file_path = info['file_path']
        self.parameters = info['parameters']
        self.progress = info['progress']
        self.info = info
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))
        with open(file_path, 'w+') as jf:
            json.dump(self.info, jf, indent=2)

    def finish_stage(self, stage):
        self.progress[stage] = 'Finish'
        self.info['progress'][stage] = 'Finish'
        with open(self.file_path, 'w') as jf:
            json.dump(self.info, jf, indent=2)

    def reset_stage(self, stage):
        self.progress[stage] = 'No'
        self.info['progress'][stage] = 'No'
        with open(self.file_path, 'w') as jf:
            json.dump(self.info, jf, indent=2)

    def check_consist(self, parameters_dict):
        if self.version != SCRIP_VERSION:
            return False
        inconsistent_key = []
        for k in self.parameters.keys():
            try:
                if self.parameters[k] == parameters_dict[k]:
                    continue
                else:
                    inconsistent_key.append(k)
            except KeyError:
                inconsistent_key.append(k)
        if inconsistent_key.__len__() != 0:
            return False
        else:
            return True

    def safe_run_and_store(self, func, func_arg_list, store_path, stage):
        if self.info['progress'][stage] == 'No':
            ret = func(*func_arg_list)
            store_to_pickle(ret, store_path)
            self.finish_stage(stage)
        else:
            ret = read_pickle(store_path)
        return ret

    def safe_run(self, func, func_arg_list, stage):
        if self.info['progress'][stage] == 'No':
            func(*func_arg_list)
            self.finish_stage(stage)
        else:
            pass
        return



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   enhance.py
@Time    :   2021/08/24 10:40:23
@Author  :   Xin Dong 
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''


from sklearn.neighbors import BallTree
from sklearn.neighbors import KDTree
import sys
from SCRIP.utilities.utils import print_log, excute_info, safe_makedirs
# from SCRIP.enrichment.bed_generation import generate_beds
# from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, wait, ALL_COMPLETED, as_completed
import scanpy as sc
import anndata as ad
import random
import scipy
import numpy as np
import pandas as pd
from SCRIP.enrichment.bed_generation import generate_peak_list
from multiprocessing import Pool


def determine_number_of_cells_per_group(input_mat, start=1, end=70, iteration_time=30, aim_peak_number=10000, peak_confidence=None):
    peak_number = 0
    cell_pool = input_mat.obs_names.to_list()
    cell_number = start - 1
    while peak_number <= aim_peak_number:
        cell_number += 1
        if peak_confidence == None:
            peak_confidence = np.ceil(0.2*cell_number)
        tmp_peak_number = []
        for i in range(iteration_time):
            random.seed(i)
            cells = random.sample(cell_pool, cell_number)
            tmp_peak_number.append(generate_peak_list(cells, input_mat, peak_confidence).__len__())
        peak_number = sum(tmp_peak_number)/iteration_time
        if cell_number > end:
            print_log('Can not find enough cells!')
            sys.exit()
    return cell_number


def sub_coor_table_in_small_square(cell, coor_table, step, t):
    tmp = coor_table[(coor_table['X'] < coor_table.loc[cell, 'X'] + step * t) &
                     (coor_table['X'] > coor_table.loc[cell, 'X'] - step * t) &
                     (coor_table['Y'] < coor_table.loc[cell, 'Y'] + step * t) &
                     (coor_table['Y'] > coor_table.loc[cell, 'Y'] - step * t)]
    return tmp.copy()


def find_nearest_cells_umap(cell, coor_table, n_neighbor=20, step=None):
    #   coor_table has two columns: X, Y, the index of coortable is cell barcodes
    if step == None:
        up_limit = coor_table.max()
        down_limit = coor_table.min()
        width = up_limit.X - down_limit.X
        height = up_limit.Y - down_limit.Y
        step = min(width, height)/200
    t = 1
    tmp = pd.DataFrame()
    while tmp.shape[0] < n_neighbor:
        tmp = sub_coor_table_in_small_square(cell, coor_table, step, t)
        t += 1
    squ_distance = [(tmp.loc[i, "X"] - tmp.loc[cell, "X"])**2 + (tmp.loc[i, "Y"] - tmp.loc[cell, "Y"])**2 for i in tmp.index]
    tmp['distance'] = squ_distance
    tmp = tmp.sort_values(by='distance', ascending=True)
    # since we use square to check the distance, check whether the point is out the circle
    while tmp.iloc[n_neighbor-1, 2] > ((step*t)**2):
        t += 1
        tmp = sub_coor_table_in_small_square(cell, coor_table, step, t)
        squ_distance = [(tmp.loc[i, "X"] - tmp.loc[cell, "X"])**2 + (tmp.loc[i, "Y"] - tmp.loc[cell, "Y"])**2 for i in tmp.index]
        tmp['distance'] = squ_distance
        tmp = tmp.sort_values(by='distance', ascending=True)
    neighbor_bcs = tmp.index[0:n_neighbor].tolist()
    return neighbor_bcs

def cal_neighbor_cell_peak_mat_umap(table, input_mat, coor_table, impute_n, i):
    # print(i)
    for cell_barcode in table.index:
        nearest_bc = find_nearest_cells_umap(cell_barcode, coor_table, n_neighbor=impute_n)
#         table.loc[cell_barcode,:] = [1 if i else 0 for i in input_mat.loc[nearest_bc,:].sum() >= 1]
        table.loc[cell_barcode, :] = input_mat.loc[nearest_bc, :].sum()
    return table

def cal_neighbor_cell_peak_mat_umap_batch(input_mat, coor_table, impute_n, n_cores=8):
    print_log("Generating neighbor cells peak matrix, divide into {n} chunks...".format(n=n_cores))
    input_table_split = np.array_split(input_mat, n_cores)
    args = [[table, input_mat, coor_table, impute_n, i] for (i, table) in enumerate(input_table_split)]
    with Pool(n_cores) as p:
        result = p.starmap(cal_neighbor_cell_peak_mat_umap, args)
    cell_peak = pd.concat([i for i in result])
    print_log('Finished!')
#     return scipy.sparse.csr_matrix(cell_peak)
    return cell_peak


def find_nearest_cells(q_point, tree, n_neighbor):
    _, ind = tree.query(q_point, k=n_neighbor+1)
    return ind[0][1:]


def cal_neighbor_cell_peak_mat(sub_mat, input_mat, tree, pc_table, impute_n, start_idx, i):
    end_index = start_idx + sub_mat.shape[0]
    k = 0
    for idx in range(start_idx, end_index):
        nearest_bc_idx = find_nearest_cells(np.reshape(pc_table[idx, :], (1, -1)), tree, n_neighbor=impute_n)
#         scipy.sparse.csr_matrix(input_mat[nearest_bc_idx,:].sum(0))
        sub_mat[k, :] = input_mat[nearest_bc_idx, :].sum(0)
        k += 1
    return sub_mat


def cal_neighbor_cell_peak_mat_batch(input_mat, impute_n=5, KD_leafsize=80, nPC=50, n_cores=8):
    '''
    input_mat:
    a csr sparse matrix, which can be get by adata.X
    
    '''
    print_log("Building KD tree...")
    pc_table = sc.tl.pca(input_mat, n_comps=50, svd_solver='arpack')
    tree = BallTree(pc_table, KD_leafsize)
    print_log("Calculating neighbors, divide into {n} chunks...".format(n=n_cores))
    cell_number = input_mat.shape[0]
    index_split = [i for i in range(0, cell_number, int(cell_number/n_cores))] + [cell_number]
#     input_table_split = np.array_split(input_mat_dense, n_cores)
    input_mat_lil = input_mat.tolil()
    input_mat_split = [input_mat_lil[index_split[i]:index_split[i+1], :] for i in range(index_split.__len__()-1)]
    args = [[sub_mat, input_mat, tree, pc_table, impute_n, index_split[i], i] for (i, sub_mat) in enumerate(input_mat_split)]
#     print(args)
    with Pool(n_cores) as p:
        result = p.starmap(cal_neighbor_cell_peak_mat, args)
    cell_peak_csr = scipy.sparse.vstack(result).tocsr()
    print_log('Finished!')
    return cell_peak_csr

# @excute_info('Start generating nearest neighbor cells beds ...', 'Finished generating nearest neighbor cells beds!')
# def generate_neighbor_bed_umap(adata, input_mat, bed_path, map_dict_store_path, peaks_number_store_path, n_neighbor=10, peak_confidence=2, n_cores=8):
#     coor_table = pd.DataFrame(adata.obsm['X_umap'], index=adata.obs.index, columns=["X", "Y"])
#     width = coor_table.max().X - coor_table.min().X
#     height = coor_table.max().Y - coor_table.min().Y
#     step = min(width, height)/200
#     map_dict = {}
#     safe_makedirs(bed_path)
#     # total_cnt = adata.obs.index.__len__()
#     executor = ThreadPoolExecutor(max_workers=n_cores)
#     all_task = []
#     for cell in adata.obs.index:
#         neighbor_cells = find_nearest_cells_umap(cell, coor_table, n_neighbor, step)
#         map_dict[cell] = neighbor_cells
#         all_task.append(executor.submit(generate_beds, bed_path + "/" + str(cell) + ".bed", neighbor_cells, input_mat, peak_confidence))
#     wait(all_task, return_when=ALL_COMPLETED)
#     pd.DataFrame([_.result() for _ in as_completed(all_task)]).to_csv(peaks_number_store_path, header=None, index=None, sep='\t')
#     with open(map_dict_store_path, "wb") as map_dict_file:
#         pickle.dump(map_dict, map_dict_file)
#     return map_dict


def enhance(input_mat, impute_n=5, KD_leafsize=80, nPC=50, path='SCRIPT/enhancement/', binarize=True, n_cores=8):
    '''
    input_mat:
    a csr sparse matrix
    
    '''
    safe_makedirs(path)
    imputed_csr = cal_neighbor_cell_peak_mat_batch(input_mat, impute_n=impute_n, KD_leafsize=KD_leafsize, nPC=nPC, n_cores=n_cores)
    if binarize == True:
        imputed_csr[imputed_csr > 1] = 1
    utils.store_to_pickle(imputed_csr, path + 'imputed.csr.pk')
    return imputed_csr

def run(args):
    processed_adata_path = args.processed_experiment
    feature_matrix_path = args.feature_matrix
    project = args.project
    cell_number_impute = args.cell_number_impute
    peak_confidence_impute = args.peak_confidence_impute
    yes = args.yes
    clean = args.clean
    n_cores = args.n_cores

    print_log('Reading Files, Please wait ...')
    processed_adata = ad.read_h5ad(processed_adata_path)
    feature_matrix = sc.read_10x_h5(feature_matrix_path, gex_only=False)

    if peak_confidence_impute == 'auto':
        if cell_number_impute == 'auto':
            print_log('Estimating the number of cells per group...')
            cell_number_impute = determine_number_of_cells_per_group(
                feature_matrix, start=1, end=70, iteration_time=30, aim_peak_number=10000, peak_confidence=None)
            print_log('Merge {cell_number} neighbor cells as one!'.format(cell_number=cell_number_impute))
        else:
            cell_number_impute = int(cell_number_impute)
        peak_confidence_impute = int(np.ceil(0.1*cell_number_impute))
    else:
        peak_confidence_impute = int(peak_confidence_impute)
        if cell_number_impute == 'auto':
            print_log('Estimating the number of cells per group...')
            cell_number_impute = determine_number_of_cells_per_group(
                feature_matrix, start=1, end=70, iteration_time=30, aim_peak_number=10000, peak_confidence=peak_confidence_impute)
            print_log('Merge {cell_number} neighbor cells as one!'.format(cell_number=cell_number_impute))
        else:
            cell_number_impute = int(cell_number_impute)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   start.py
@Time    :   2021/04/16 12:35:34
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import argparse
import sys
import ruamel.yaml
from SCRIP.Constants import SCRIP_VERSION
from SCRIP.enrichment.enrich import run_enrich
from SCRIP.imputation.impute import run_impute
from SCRIP.targets.target import run_target
from SCRIP.conf.config import update_setting
from SCRIP.utilities.utils import read_config
yaml = ruamel.yaml.YAML()

CONFIG, CONFIG_PATH = read_config()

def main():
    argparser = prepare_argparser()
    args = argparser.parse_args()

    subcommand  = args.subcommand

    if subcommand == "enrich":
        try:
            run_enrich(args)
        except MemoryError:
            sys.exit( "MemoryError occurred.")
    elif subcommand == "impute":
        try:
            run_impute(args)
        except MemoryError:
            sys.exit("MemoryError occurred.")
    elif subcommand == "target":
        try:
            run_target(args)
        except MemoryError:
            sys.exit("MemoryError occurred.")
    elif subcommand == "config":
        try:
            update_setting( args )
        except:
            sys.exit("Setting set wrong.")
    elif subcommand == "format":
        pass

    return

def prepare_argparser():
    description = "%(prog)s"
    epilog = "For command line options of each command, type: %(prog)s COMMAND -h"

    # top-level parser
    argparser = argparse.ArgumentParser( description = description, epilog = epilog )
    argparser.add_argument( "--version", action="version", version="%(prog)s "+SCRIP_VERSION )
    subparsers = argparser.add_subparsers( dest = 'subcommand' )
    subparsers.required = True

    add_enrich_parser(subparsers)
    add_impute_parser(subparsers)
    add_target_parser(subparsers)
    add_config_parser(subparsers)

    return argparser


def add_config_parser(subparsers):
    argparser_setting = subparsers.add_parser("config", help="Configuration.")
    argparser_setting.add_argument( "--show", dest = "show", action = 'store_true', default=False, help = "" )
    argparser_setting.add_argument("--human_tf_index", dest="human_tf_index", type=str, help="")
    argparser_setting.add_argument("--human_hm_index", dest="human_hm_index", type=str, help="")
    argparser_setting.add_argument("--mouse_tf_index", dest="mouse_tf_index", type=str, help="")
    argparser_setting.add_argument("--mouse_hm_index", dest="mouse_hm_index", type=str, help="")


def add_enrich_parser( subparsers ):
    """Add main function 'enrich' argument parsers.
    """
    argparser_enrich = subparsers.add_parser("enrich", help="Main function.")

    # group for input files
    group_input = argparser_enrich.add_argument_group( "Input files arguments" )
    group_input.add_argument( "-i", "--input_feature_matrix", dest = "feature_matrix", type = str, required = True,
                              help = 'A cell by peak matrix . REQUIRED.' )
    group_input.add_argument( "-s", "--species", dest = "species", choices= ['hs', 'mm'], required = True,
                              help = 'Species. "hs"(human) or "mm"(mouse). REQUIRED.' )
    # group for output files
    group_output = argparser_enrich.add_argument_group( "Output arguments" )
    group_output.add_argument( "-p", "--project", dest = "project", type = str, default = "" ,
                               help = 'Project name, which will be used to generate output files folder. DEFAULT: Random generate.')

    # group for preprocessing
    group_preprocessing = argparser_enrich.add_argument_group( "Preprocessing paramater arguments" )
    group_preprocessing.add_argument( "--min_cells", dest = "min_cells", type = str, default = 'auto',
                                      help='Minimal cell cutoff for features. Auto will take 0.05%% of total cell number.DEFAULT: "auto".')
    group_preprocessing.add_argument("--min_peaks", dest="min_peaks", type=str, default='auto',
                                     help='Minimal peak cutoff for cells. Auto will take the mean-3*std of all feature number (if less than 500 is 500). DEFAULT: "auto".')
    group_preprocessing.add_argument("--max_peaks", dest="max_peaks", type=str, default='auto',
                                     help='Max peak cutoff for cells. This will help you to remove the doublet cells. Auto will take the mean+5*std of all feature number. DEFAULT: "auto".')

    group_other = argparser_enrich.add_argument_group( "Other options" )
    group_other.add_argument( "-t", '--thread', dest='n_cores', type = int, default = 16,
                              help="Number of cores use to run SCRIP. DEFAULT: 16.")
    group_other.add_argument( "-y", "--yes", dest='yes', action = 'store_true', default = False,
                              help="Whether ask for confirmation. DEFAULT: False.")
    group_other.add_argument( "--clean", dest = "clean", action = 'store_true', default = False,
                              help="Whether delete tmp files(including bed and search results) generated by SCRIP. DEFAULT: False.")


def add_impute_parser(subparsers):
    """Add main function 'impute' argument parsers.
    """
    argparser_impute = subparsers.add_parser("impute", help="Imputation Factor function.")

    # group for input files
    group_input = argparser_impute.add_argument_group("Input files arguments")
    group_input.add_argument("-i", "--input_feature_matrix", dest="feature_matrix", type=str, required=True,
                             help='A cell by peak matrix. h5 or h5ad supported. REQUIRED.')
    group_input.add_argument("-s", "--species", dest="species", choices=['hs', 'mm'], required=True,
                             help='Species. "hs"(human) or "mm"(mouse). REQUIRED.')

    # group for output files
    group_output = argparser_impute.add_argument_group("Output arguments")
    group_output.add_argument("-p", "--project", dest="project", type=str, default="",
                              help='Project name, which will be used to generate output files folder. DEFAULT: Random generate.')
    group_output.add_argument("-f", "--format", dest="format", choices=['h5ad', 'mtx'], default="h5ad",
                              help='Format generate for output RP count. DEFAULT: h5ad.')

    # group for impute
    group_impute = argparser_impute.add_argument_group("Peak imputation paramater arguments")
    group_impute.add_argument("--factor", dest="factor", type=str, default='', required=True,
                              help='The factor you want to impute. REQUIRED.')
    group_impute.add_argument("--ref_baseline", dest="ref_baseline", type=int, default=500,
                              help='Remove dataset which peaks number less than this value. DEFAULT: 500.')
    group_impute.add_argument("--remove_others", dest="remove_others", action='store_true',  default=False,
                              help='Remove signal not from best match. DEFAULT: False.')

    group_other = argparser_impute.add_argument_group("Other options")
    group_other.add_argument("--min_cells", dest="min_cells", type=str, default='auto',
                                     help='Minimal cell cutoff for features. Auto will take 0.05%% of total cell number.DEFAULT: "auto".')
    group_other.add_argument("--min_peaks", dest="min_peaks", type=str, default='auto',
                                     help='Minimal peak cutoff for cells. Auto will take the mean-3*std of all feature number (if less than 500 is 500). DEFAULT: "auto".')
    group_other.add_argument("--max_peaks", dest="max_peaks", type=str, default='auto',
                                     help='Max peak cutoff for cells. This will help you to remove the doublet cells. Auto will take the mean+5*std of all feature number. DEFAULT: "auto".')
    group_other.add_argument("-t", '--thread', dest='n_cores', type=int, default=16,
                             help="Number of cores use to run SCRIP. DEFAULT: 16.")


def add_target_parser(subparsers):
    """Add main function 'target' argument parsers.
    """
    argparser_impute = subparsers.add_parser("target", help="Calculate targets based on factor peak count.")

    # group for input files
    group_input = argparser_impute.add_argument_group("Input files arguments")
    group_input.add_argument("-i", "--input_feature_matrix", dest="feature_matrix", type=str, required=True,
                             help='A cell by peak matrix. h5 or h5ad supported. REQUIRED.')
    group_input.add_argument("-s", "--species", dest="species", choices=['hs', 'mm'], required=True,
                             help='Species. "hs"(human) or "mm"(mouse). REQUIRED.')

    # group for output files
    group_output = argparser_impute.add_argument_group("Output arguments")
    group_output.add_argument("-o", "--output", dest="output", type=str, default="RP.h5ad",
                              help='output h5ad file. DEFAULT: RP.h5ad')

    group_other = argparser_impute.add_argument_group("Other options")
    group_other.add_argument("-d", '--decay', dest='decay', type=int, default=10000,
                             help="Range to the effect of peaks. DEFAULT: 10000.")

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.stderr.write("User interrupted!\n")
        sys.exit(0)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   validation.py
@Time    :   2021/04/16 12:34:48
@Author  :   Xin Dong 
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

from SCRIP.utilities.utils import print_log
import re
import os
import sys


def check_para(cell_feature_adata, project,
               chip_index,  
               beds_path, chip_result_path, result_store_path,
               yes, clean, n_cores):
    # if processed_adata.shape[0] > cell_feature_adata.shape[0]:
    #     print_log('WARNING: There are more cells in anndata than input feature matrix.')
    if re.match(r'chr.*_\d*_\d*', cell_feature_adata.var_names[0]) == None:
        print_log("cell_feature_adata's index/rownames should be like this: chr1_222222_333333. Each row is a feature and each column is a cell.")
        sys.exit()
    # if reference_method in ['chip' , 'integration' , 'both']:
    #     if not os.path.exists(chip_index) or chip_index == '':
    #         print_log("chip_index do not exist!")
    #         sys.exit()
    # if reference_method in ['motif', 'integration', 'both']:
    #     if not os.path.exists(motif_index) or motif_index == '':
    #         print_log("motif_index do not exist!")
    #         sys.exit()
    # prepare information
    information = '\n~~~~~\nWelcome to use SCRIP. Here are your settings:\n\n'
    information += 'There are UMAP information and cluster information in your processed_adata. '
    information += 'The whole output file will be stored in {project}. '.format(project=project)
    # information += 'This will borrow peaks from {cell_number_impute} nearest cells for every cell. '.format(cell_number_impute=cell_number_impute)
    # information += 'For each cell, peaks that confidence is less than {peak_confidence_impute} will be deleted.\n'.format(peak_confidence_impute=peak_confidence_impute)
    information += 'The peaks will be stored in {beds_path}.\n'.format(beds_path=beds_path)
    # if reference_method == 'integration' or reference_method == 'both':
    #     information += 'SCRIP will search ChIP-seq and motif index both.\n'
    #     if reference_method == 'integration':
    #         information += 'After that, SCRIP will integrate into one result based on variability from each result.\n'
    # elif reference_method == 'chip':
    #     information += 'SCRIP will search ChIP-seq index.\n'
    # elif reference_method == 'motif':
    #     information += 'SCRIP will search motif index.\n'
    # if reference_method != 'motif':
    #     information += 'ChIP-seq index locates at {chip_index}.\n'.format(chip_index=chip_index)
    #     information += 'ChIP-seq intervals intersection result will be stored in {chip_result_path}'.format(chip_result_path=chip_result_path)
    # if reference_method != 'chip':
    #     information += 'Motif index locates at {motif_index}.\n'.format(motif_index=motif_index)
    #     information += 'Motif intervals intersection result will be stored in {motif_result_path}, '.format(motif_result_path=motif_result_path)
    if result_store_path != '':
        information += 'Computed result will be stored in {result_store_path}.\n'.format(result_store_path=result_store_path)
    else:
        information += 'Computed result will not be stored.\n'
    information += 'All folders will{not_string} be removed after processing. '.format(not_string='' if clean == True else ' not')
    information += 'All processes will use {n_cores} cores.\n~~~~~\n'.format(n_cores=n_cores)
    print(information)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   convert_format.py
@Time    :   2021/04/16 12:35:19
@Author  :   Xin Dong 
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import scanpy as sc
import os
import sys
import anndata as ad
import h5py


def read_SingleCellExperiment_rds(input_RDS):
    import anndata2ri
    from rpy2.robjects import r
    anndata2ri.activate()
    rscript = 'readRDS("{RDS_file_path}")'.format(RDS_file_path = input_RDS)
    adata = r(rscript)
    adata.var.columns = [str(i) for i in adata.var.columns]
    adata.obs.columns = [str(i) for i in adata.obs.columns]
    return adata


#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   calculation.py
@Time    :   2021/04/16 12:35:01
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import numpy as np
import pandas as pd
# import scipy
from sklearn import preprocessing
from SCRIP.utilities.utils import excute_info
# from multiprocessing import Process, Pool


def standardScaler(t):
    scaler = preprocessing.StandardScaler().fit(t)
    rt = pd.DataFrame(scaler.transform(t), index=t.index, columns=t.columns)
    return rt


@excute_info('Summary result from dataset level to factor level.')
def map_factor_on_ChIP(table):
    ret_table = table.copy()
    # map factor by id "_"
    factor_index_list = []
    for i in ret_table.index:
        factor_name = i.split("_")
        factor_index_list.append(factor_name[0])
    ret_table.loc[:, "Factor"] = factor_index_list
    factor_table = ret_table.groupby("Factor").max()
    return factor_table

@excute_info('Getting the best reference for each cell.')
def get_factor_source(table):
    ret_table = table.copy()
    # map factor by id "_"
    factor_index_list = []
    for i in ret_table.index:
        factor_name = i.split("_")
        factor_index_list.append(factor_name[0])
    ret_table.loc[:, "Factor"] = factor_index_list
    max_index = ret_table.groupby("Factor").idxmax()
    return max_index


def cal_score(dataset_overlap_df, peaks_number, qpeak_length):
    '''
    nql: normalize query peak length
    dm: minus the mean
    '''
    dataset_cell_percent = (dataset_overlap_df.T/peaks_number.loc[dataset_overlap_df.index, 1]).T
    dataset_cell_percent_dl = dataset_cell_percent/qpeak_length.loc[dataset_cell_percent.columns, 1]
    dataset_cell_percent_dl_dm = (dataset_cell_percent_dl.T - dataset_cell_percent_dl.mean(1)).T
    # dataset_cell_percent_scale = (dataset_cell_percent/dataset_cell_percent.sum())*1e4
    # dataset_cell_percent_scale_dm = (dataset_cell_percent_scale.T/dataset_cell_percent_scale.mean(1)).T
    return dataset_cell_percent_dl_dm


# def adjust_outlier(mat, r=5):
#     '''
#     columns are cells
#     rows are factors
#     '''
#     ret = mat.copy()
#     max_v = (mat.mean(1)+r*mat.std(1))
#     min_v = (mat.mean(1)-r*mat.std(1))
#     for i in mat.index:
#         ret.loc[i, ret.columns[ret.loc[i, :] > max_v[i]]] = max_v[i]
#         ret.loc[i, ret.columns[ret.loc[i, :] < min_v[i]]] = min_v[i]
#     return ret

def score_normalization(dataset_cell_df):
    tf_cell_df = map_factor_on_ChIP(dataset_cell_df)
    tmp = standardScaler(tf_cell_df.T).T
    tf_cell_df_lsn = 1/(1+np.exp(-tmp))  # LSN(Logistic Sigmoid Normalisation)
    tf_cell_df_lsn_std = standardScaler(tf_cell_df_lsn)
    return tf_cell_df_lsn_std



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   convert_seurat2anndata.py
@Time    :   2021/04/16 12:35:25
@Author  :   Xin Dong 
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import scanpy as sc
# import os
# import sys
# import numpy as np
# import pandas as pd
# import matplotlib.pyplot as plt
# import sklearn
import anndata2ri
from rpy2.robjects import r
from rpy2.robjects.conversion import localconverter
# #Loading the rpy2 extension enables cell magic to be used
# #This runs R code in jupyter notebook cells
# %load_ext rpy2.ipython

# sc.settings.verbosity = 3
# sc.logging.print_versions()

rscript = '''
obj <- readRDS('/path/to/file.rds')
Seurat::as.SingleCellExperiment(obj$ATAC)
'''

#rscript = "print(1)"
with localconverter(anndata2ri.converter):
    adata = r(rscript)

print(adata)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   search_giggle.py
@Time    :   2021/04/16 12:34:35
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''


import subprocess
import os
from multiprocessing import Pool
import warnings
# from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, wait, ALL_COMPLETED
import pandas as pd
import numpy as np
from SCRIP.utilities.utils import print_log, safe_makedirs

warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)


def search_ref(bed_path, result_path, index_path):
    cmd = f'giggle search -i "{index_path}" -q "{bed_path}" -s > "{result_path}"\n'
    # cmd = f'igd search {index_path}/ref.igd -q {bed_path} | head -n -1 | cut -f 2,3,4 > {result_path}'
    # cmd = f'seqpare "{index_path}/*.bed.gz" "{bed_path}" -m 1 -o {result_path}\n'
    subprocess.run(cmd, shell=True, check=True)

def search_ref_batch(bed_folder, result_folder, index_path, n_cores=8, tp=''):
    print_log(f'Start searching beds from {tp} index ...')
    safe_makedirs(result_folder)
    beds = os.listdir(bed_folder)
    args = []
    for bed in beds:
        barcodes = bed[:-7]  # remove suffix '.bed.gz'
        args.append((os.path.join(bed_folder, bed),
                     os.path.join(result_folder, barcodes + '.txt'),
                     index_path))
    with Pool(n_cores) as p:
        p.starmap(search_ref, args)
    print_log(f'Finished searching beds from {tp} index ...')


def read_search_result(files):
    for i in enumerate(files):
        result_name = os.path.basename(i[1])
        cell_bc = result_name[:-4]  # remove suffix '.txt'
        dtframe = pd.read_csv(i[1], sep="\t", index_col=0, comment='#', header=None)
        read_col = 2  # 1 file_size 2 overlaps 3 odds_ratio 4 fishers_two_tail 5 fishers_left_tail 6 fishers_right_tail 7 combo_score
        if i[0] == 0:
            dtframe = dtframe.loc[:, [read_col]].copy()
            dataset_cell_score_df = dtframe.rename(columns={read_col: cell_bc}).copy()
        else:
            dataset_cell_score_df[cell_bc] = dtframe.loc[:, read_col]
    dataset_cell_score_df.index = [n.rsplit('/', 1)[0][:-7] for n in dataset_cell_score_df.index]  # remove suffix '.bed.gz'
    return dataset_cell_score_df


def read_search_result_batch(path, n_cores=8, tp=''):
    print_log(f"Reading searching results, using {n_cores} cores...")
    file_list = os.listdir(path)
    result_split = np.array_split(file_list, n_cores)
    args = [[[os.path.join(path, j) for j in list_chunk]] for list_chunk in result_split]
    with Pool(n_cores) as p:
        result = p.starmap(read_search_result, args)
    dataset_cell_score_df = pd.concat([i for i in result], axis=1)
    print_log(f"Finished reading {tp} index search result!")
    return dataset_cell_score_df



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   impute.py
@Time    :   2021/08/25 19:09:37
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''
import os
import subprocess
import random
import time
# from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, wait, ALL_COMPLETED, as_completed
from multiprocessing import Pool
import pandas as pd
import scipy
import pybedtools
import scanpy as sc
from SCRIP.enrichment.search import read_search_result_batch
from SCRIP.enrichment.bed_generation import generate_beds_by_matrix
from SCRIP.utilities.utils import print_log, safe_makedirs, excute_info, write_to_mtx, store_to_pickle, read_config




def search_ref_factor(bed_path, result_path, index_path, factor):
    cmd = f'giggle search -i "{index_path}" -q "{bed_path}" -s -f {factor}_ > "{result_path}"\n'
    # cmd = f'igd search {index_path}/ref.igd -q {bed_path} | head -n -1 | cut -f 2,3,4 > {result_path}'
    # cmd = f'seqpare "{index_path}/*.bed.gz" "{bed_path}" -m 1 -o {result_path}\n'
    subprocess.run(cmd, shell=True, check=True)

def search_ref_factor_batch(bed_folder, result_folder, index_path, factor, n_cores=8, tp=''):
    print_log(f'Start searching beds from {tp} index ...')
    safe_makedirs(result_folder)
    beds = os.listdir(bed_folder)
    args = []
    for bed in beds:
        barcodes = bed[:-7]  # remove suffix '.bed.gz'
        args.append((os.path.join(bed_folder, bed),
                     os.path.join(result_folder, barcodes + '.txt'),
                     index_path,
                     factor))
    with Pool(n_cores) as p:
        p.starmap(search_ref_factor, args)
    print_log(f'Finished searching beds from {tp} index ...')

# def get_factor_affinity(input_mat, bed_file_path, reference, factor, ccre_number):
#     peaks = input_mat.var_names.to_list()
#     peaks_number = peaks.__len__()
#     ref_number = pd.read_csv(reference + '/peaks_number.txt', sep='\t', index_col=0, header=None)
#     factor_idx = [i for i in ref_number.index if i.startswith(factor)]
#     ref_number = ref_number.loc[factor_idx,:]
#     peaks = pd.DataFrame([p.rsplit("_", 2) for p in peaks])
#     peaks.to_csv(bed_file_path, sep="\t", header=None, index=None)
#     cmd = 'sort --buffer-size 2G -k1,1 -k2,2n -k3,3n {bed_path} | bgzip -c > {bed_path}.gz\n'.format(bed_path=bed_file_path)
#     cmd += 'rm {bed_path}'.format(bed_path=bed_file_path)
#     subprocess.run(cmd, shell=True, check=True)

#     search_seqpare_factor(bed_file_path + '.gz', bed_file_path[0:-4] + '.txt', reference, factor)
#     all_peak_result = read_seqpare_result([bed_file_path[0:-4] + '.txt'])
#     affinity = all_peak_result.iloc[:,0]/(ref_number[1]*peaks_number/ccre_number)
#     return affinity


@excute_info('Getting the best reference for each cell.')
def get_factor_source(table):
    ret_table = table.copy()
    # map factor by id "_"
    factor_index_list = []
    for i in ret_table.index:
        factor_name = i.split("_")
        factor_index_list.append(factor_name[0])
    ret_table.loc[:, "Factor"] = factor_index_list
    max_index = ret_table.groupby("Factor").idxmax()
    return max_index

def impute(input_mat_adata, impute_factor, ref_path, bed_check=True, search_check=True, path='SCRIP/imputation/', write_format='', ref_baseline=500, remove_others_source=False, n_cores=8):
    '''
    impute tf ChIP data from ATAC data
    '''

    safe_makedirs(path)
    print(input_mat_adata.X.shape)
    if bed_check == True:
        if not os.path.exists(f'{path}/imputed_beds/'):
            print_log('Generating beds...')
            generate_beds_by_matrix(input_mat_adata, f'{path}/imputed_beds/', f'{path}/imputed_beds_peaks_number.txt', n_cores)
        else:
            print_log('Skip generate beds...')
    else:
        print_log('Generating beds...')
        generate_beds_by_matrix(input_mat_adata, f'{path}/imputed_beds/', f'{path}/imputed_beds_peaks_number.txt', n_cores)

    if search_check == True:
        if not os.path.exists(f'{path}/imputed_results_{impute_factor}/'):
            search_ref_factor_batch(f'{path}/imputed_beds/', f'{path}/imputed_results_{impute_factor}/', ref_path, impute_factor, n_cores)
        else:
            print_log('Skip searching beds...')
    else:
        search_ref_factor_batch(f'{path}/imputed_beds/', f'{path}/imputed_results_{impute_factor}/', ref_path, impute_factor, n_cores)

    print_log('Calculating score...')
    factor_enrich = read_search_result_batch(f'{path}/imputed_results_{impute_factor}/', n_cores)

    peaks_number = pd.read_csv(os.path.join(ref_path, 'peaks_number.txt'), sep='\t', header=None, index_col=0)
    peaks_number_factor = peaks_number.loc[[i for i in peaks_number.index if i.startswith(f'{impute_factor}_')], :].copy()
    peaks_number_baseline_index = peaks_number_factor.index[peaks_number_factor[1] > ref_baseline]
    # peaks_number_factor = peaks_number_factor.loc[peaks_number_baseline_index, :]

    factor_enrich = factor_enrich.loc[peaks_number_baseline_index, :].copy()
    factor_score = (factor_enrich.T/peaks_number_factor.loc[factor_enrich.index, 1]).T

    factor_source = get_factor_source(factor_score)
    store_to_pickle(factor_source, f'{path}/{impute_factor}_dataset_source.pk')

    chip_bed_list = [pybedtools.BedTool(os.path.join(ref_path, 'raw_beds', i + '.bed.gz')) for i in factor_source.iloc[0, :].unique()]
    if len(chip_bed_list) == 1:
        chip_bed = chip_bed_list[0]
    else:
        chip_bed = chip_bed_list[0].cat(*chip_bed_list[1:])
    data_bed = pybedtools.BedTool('\n'.join(['\t'.join(p.rsplit('_', maxsplit=2)) for p in input_mat_adata.var_names]), from_string=True)
    intersect_bed = data_bed.intersect(chip_bed, u=True)
    imputed_chip_peak = str(intersect_bed).replace('\t', '_').split('\n')[0:-1]

    chip_cell_peak = input_mat_adata[:, imputed_chip_peak].copy()
    chip_cell_peak_df = chip_cell_peak.to_df()
    if remove_others_source == True:
        for i in factor_source.iloc[0, :].unique():
            cellbc = factor_source.columns[factor_source.iloc[0, :] == i]
            tmp_dataset_bed = pybedtools.BedTool(os.path.join(ref_path, 'raw_beds', i + '.bed.gz'))
            exclude_chip_peak = str(intersect_bed.intersect(tmp_dataset_bed, v=True)).replace('\t', '_').split('\n')[0:-1]
            chip_cell_peak_df.loc[cellbc, exclude_chip_peak] = 0
    chip_cell_peak = sc.AnnData(chip_cell_peak_df)
    chip_cell_peak.X = scipy.sparse.csr.csr_matrix(chip_cell_peak.X)
    if write_format != '':
        print_log('Writing results...')
        if write_format == 'h5ad':
            chip_cell_peak.write_h5ad(f'{path}/imputed_{impute_factor}.h5ad')
        elif write_format == 'mtx':
            write_to_mtx(chip_cell_peak, f'{path}/imputed_{impute_factor}_mtx/')
    print_log('Finished!')
    return chip_cell_peak


def run_impute(args):
    feature_matrix_path = args.feature_matrix
    species = args.species
    factor = args.factor
    project = args.project
    write_format = args.format
    min_cells = args.min_cells  # for removing few features cells
    min_peaks = args.min_peaks  # for removing few cells features
    max_peaks = args.max_peaks  # for removing doublet cells
    remove_others = args.remove_others
    ref_baseline = args.ref_baseline
    n_cores = args.n_cores

    if project == '':
        tmp_chr_list = [chr(i) for i in range(ord("A"), ord("Z") + 1)] + [chr(i) for i in range(ord("a"), ord("z") + 1)] + [chr(i) for i in range(ord("0"), ord("9") + 1)]
        random.seed(time.time())
        tmp_prefix = str(time.time())[6:13].replace('.', '') + '_' + ''.join(random.sample(tmp_chr_list, 4))
        project = 'SCRIP_' + tmp_prefix

    if factor in ['H3K4me3', 'H3K4me2', 'H3K27ac', 'H3K9ac', 'H3K4me1']:
        factor_type='histone'
    else:
        factor_type='TR'

    CONFIG, _ = read_config()
    if factor_type== 'TR':
        if species == 'hs':
            chip_index = CONFIG['index']['human_tf_index']
        elif species == 'mm':
            chip_index = CONFIG['index']['mouse_tf_index']
        else:
            pass
    else:
        if species == 'hs':
            chip_index = CONFIG['index']['human_hm_index']
        elif species == 'mm':
            chip_index = CONFIG['index']['mouse_hm_index']
        else:
            pass

    if feature_matrix_path.endswith('.h5'):
        feature_matrix = sc.read_10x_h5(feature_matrix_path, gex_only=False)
    elif feature_matrix_path.endswith('.h5ad'):
        feature_matrix = sc.read_h5ad(feature_matrix_path)

    sc.pp.calculate_qc_metrics(feature_matrix, percent_top=None, log1p=False, inplace=True)
    feature_mean = feature_matrix.obs.n_genes_by_counts.mean()
    feature_std = feature_matrix.obs.n_genes_by_counts.std()

    if min_cells == 'auto':
        min_cells = int(0.005 * feature_matrix.n_obs)
    else:
        min_cells = int(min_cells)
    sc.pp.filter_genes(feature_matrix, min_cells=min_cells)

    if min_peaks == 'auto':
        min_peaks = 500 if feature_mean-3*feature_std < 500 else feature_mean-3*feature_std
    else:
        min_peaks = int(min_peaks)
    sc.pp.filter_cells(feature_matrix, min_genes=min_peaks)  # filter by n_genes_by_counts

    if max_peaks == 'auto':
        max_peaks = feature_mean+3*feature_std
        sc.pp.filter_cells(feature_matrix, max_genes=max_peaks)
        # feature_matrix = feature_matrix[feature_matrix.obs.n_genes_by_counts < max_peaks, :]
    elif max_peaks == 'none':
        print_log('Skip filter by max peaks number.')
    else:
        max_peaks = int(max_peaks)
        sc.pp.filter_cells(feature_matrix, max_genes=max_peaks)
        # feature_matrix = feature_matrix[feature_matrix.obs.n_genes_by_counts < max_peaks, :]

    impute(input_mat_adata=feature_matrix,
            impute_factor=factor,
            ref_path=chip_index,
            bed_check=True,
            search_check=True,
            path=f'{project}/imputation/',
            write_format=write_format,
            ref_baseline=ref_baseline,
            remove_others_source=remove_others,
            n_cores=n_cores)
    return



import os
import sys
import ruamel.yaml
from SCRIP.utilities.utils import read_config


def update_setting( args ):
    yaml = ruamel.yaml.YAML()
    yaml.default_flow_style = False
    CONFIG, CONFIG_PATH = read_config()
    if args.show:
        sys.stdout.write('The reference indices you set:\n')
        yaml.dump(CONFIG['index'], sys.stdout)
        sys.stdout.write('\n')
    else:
        if args.human_tf_index:
            CONFIG['index']['human_tf_index'] = os.path.abspath(args.human_tf_index)
        if args.human_hm_index:
            CONFIG['index']['human_hm_index'] = os.path.abspath(args.human_hm_index)
        if args.mouse_tf_index:
            CONFIG['index']['mouse_tf_index'] = os.path.abspath(args.mouse_tf_index)
        if args.mouse_hm_index:
            CONFIG['index']['mouse_hm_index'] = os.path.abspath(args.mouse_hm_index)
        with open(os.path.join(CONFIG_PATH, 'config.yml'), 'w+') as config_file:
            yaml.dump(CONFIG, config_file)
        print('Update sucessfully!')



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   bed_generation.py
@Time    :   2021/04/16 12:34:55
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import os
import subprocess
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, as_completed

import pandas as pd
import scanpy as sc

from SCRIP.utilities.utils import print_log, excute_info, safe_makedirs


def generate_peak_list(cells, input_mat, peak_confidence=1):
    cell_above_cutoff_index = sc.pp.filter_genes(
        input_mat[cells, :], min_cells=peak_confidence, inplace=False)[0]
    peaks = input_mat.var_names[cell_above_cutoff_index].to_list()
    return peaks

def generate_beds(file_path, cells, input_mat, peak_confidence=1):
    peaks = generate_peak_list(cells, input_mat, peak_confidence)
    cell_barcode = os.path.basename(file_path)[:-4]# remove .bed
    if peaks.__len__() == 0:
        print_log(f'Warning: No peaks in {file_path[:-4]}, skip generation.')
        peaks_length = 0
    else:
        peaks = pd.DataFrame([p.rsplit("_", 2) for p in peaks])
        peaks.to_csv(file_path, sep="\t", header= None, index=None)
        cmd = f'sort --buffer-size 2G -k1,1 -k2,2n -k3,3n {file_path} | bgzip -c > {file_path}.gz\n'
        cmd += f'rm {file_path}'
        subprocess.run(cmd, shell=True, check=True)
        peaks_length = peaks[2].astype(int).sum() - peaks[1].astype(int).sum()
    return [cell_barcode, peaks_length]


@excute_info('Start generating cells beds ...', 'Finished generating cells beds!')
def generate_beds_by_matrix(cell_feature_adata, beds_path, peaks_number_path, n_cores):
    safe_makedirs(beds_path)
    # total_cnt = adata.obs.index.__len__()
    executor = ThreadPoolExecutor(max_workers=n_cores)
    all_task = []
    for cell in cell_feature_adata.obs.index:
        # neighbor_cells = find_nearest_cells(cell, coor_table, n_neighbor, step)
        # map_dict[cell] = neighbor_cells
        all_task.append(executor.submit(generate_beds, beds_path + "/" + str(cell) + ".bed", cell, cell_feature_adata))
    wait(all_task, return_when=ALL_COMPLETED)
    pd.DataFrame([_.result() for _ in as_completed(all_task)]).to_csv(peaks_number_path, header=None, index=None, sep='\t')
    return



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   enrich.py
@Time    :   2021/04/16 12:34:09
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import os
import sys
import time
import shutil
import random
import scanpy as sc
import pandas as pd
from SCRIP.enrichment.bed_generation import generate_beds_by_matrix
from SCRIP.enrichment.validation import check_para
from SCRIP.enrichment.utils import EnrichRunInfo, time_estimate
from SCRIP.enrichment.calculation import score_normalization, cal_score, get_factor_source
from SCRIP.enrichment.search import search_ref_batch, read_search_result_batch
from SCRIP.utilities.utils import read_config, print_log, safe_makedirs
# from SCRIP.enhancement.enhance import determine_number_of_cells_per_group
# from SCRIP.Constants import *


def search_and_read_result(run_info, beds_path, result_path, index, n_cores):
    folder_prefix = run_info.info['project_folder']
    qpeak_length_path = os.path.join(folder_prefix, 'qpeaks_length.txt')
    qpeak_length = pd.read_csv(qpeak_length_path, sep='\t', header=None, index_col=0)/1e8
    peaks_number_path = os.path.join(index, 'peaks_number.txt')
    peaks_number = pd.read_csv(peaks_number_path, sep='\t', header=None, index_col=0)

    run_info.safe_run(search_ref_batch, [beds_path, result_path, index, n_cores], 'bed_search')
    dataset_overlap_df = run_info.safe_run_and_store(
        read_search_result_batch, [result_path, n_cores],
        os.path.join(folder_prefix, 'dataset_overlap_df.pk'),
        'dataset_overlap_df_store')
    dataset_cell_norm_df = run_info.safe_run_and_store(
        cal_score, [dataset_overlap_df, peaks_number, qpeak_length],
        os.path.join(folder_prefix, 'dataset_cell_norm_df.pk'),
        'dataset_cell_norm_df_store')
    dataset_score_source_df = run_info.safe_run_and_store(
        get_factor_source, [dataset_cell_norm_df],
        os.path.join(folder_prefix, 'dataset_score_source_df.pk'),
        'dataset_score_source_df_store')
    tf_cell_score_df = run_info.safe_run_and_store(
        score_normalization, [dataset_cell_norm_df],
        os.path.join(folder_prefix, 'tf_cell_score_df.pk'),
        'tf_cell_score_df_store')
    # transpose is used to better merge table to h5ad (anndata.obs's row is cell, col is variable)
    cell_tf_score_df = tf_cell_score_df.T
    return cell_tf_score_df, run_info, dataset_score_source_df


def enrich(cell_feature_adata, species='NA',
           project='',
           chip_index='',
           store_result_adata=True,
           n_cores=8,
           yes=False,
           clean=True):

    if project == '':
        tmp_chr_list = [chr(i) for i in range(ord("A"), ord("Z") + 1)] + [chr(i) for i in range(ord("a"), ord("z") + 1)] + [chr(i) for i in range(ord("0"), ord("9") + 1)]
        random.seed(time.time())
        tmp_prefix = str(time.time())[6:13].replace('.','') + '_' + ''.join(random.sample(tmp_chr_list, 4))
        project = 'SCRIP_' + tmp_prefix

    safe_makedirs(project)
    project_abs_path = os.path.abspath(project)
    enrich_abs_path = os.path.join(project_abs_path, 'enrichment')
    safe_makedirs(enrich_abs_path)
    ##################################
    ### check running information
    ##################################
    params = {
        'project':project_abs_path,
        'chip_index': chip_index,
        'store_result_adata': store_result_adata,
        'clean': clean,
        'species': species
    }
    run_info = EnrichRunInfo(os.path.join(enrich_abs_path, 'run_info.json'), params)
    if not run_info.check_consist(params):
        print('Not consist! Remove the folder and continue? (Y/N)')
        while True:
            confirm_info = input()
            if confirm_info == 'y' or confirm_info == 'Y':
                shutil.rmtree(enrich_abs_path)
                safe_makedirs(enrich_abs_path)
                run_info = EnrichRunInfo(os.path.join(enrich_abs_path, 'run_info.json'), params)
                break
            elif confirm_info == 'N' or confirm_info == 'n':
                print_log('Good Bye!')
                sys.exit(0)
            else:
                print('Please type Y / N.')

    beds_path = os.path.join(project, 'enrichment', 'beds')
    # total_peaks_path = os.path.join(project, 'enrichment', 'all_peaks.bed')
    qpeaks_length_path = os.path.join(project, 'enrichment', 'qpeaks_length.txt')
    chip_result_path = os.path.join(project, 'enrichment', 'ChIP_result')
    result_store_path = os.path.join(project, 'enrichment', 'SCRIP_enrichment.txt')
    safe_makedirs(beds_path)

    if species == 'hs':
        # ccre_number = 926535
        chip_factor_number = 4499
    if species == 'mm':
        # ccre_number = 339815
        chip_factor_number = 2185
    ##################################
    ### pre-check
    ##################################
    print_log('Checking parameters ...')
    check_para(cell_feature_adata, project,
               chip_index,
               beds_path, chip_result_path, result_store_path,
               yes, clean, n_cores)
    print_log('Estimating running time ...')
    elapse, future_time = time_estimate(cell_number = cell_feature_adata.shape[0],
                                        core=n_cores, chip_factor_number=chip_factor_number)
    print_log(f"It will take about {elapse} to process and finish at {future_time}.\n")

    if yes == False:
        print('Type "Y" to continue processing, "N" to abort.')
        while True:
            confirm_info = input()
            if confirm_info == 'y' or confirm_info == 'Y':
                break
            elif confirm_info == 'N' or confirm_info == 'n':
                print_log('Good Bye!')
                sys.exit(0)
            else:
                print('Please type Y / N.')
    ##################################
    ### bed generation
    ##################################
    if run_info.info['progress']['bed_generation'] == 'No':
        if os.path.exists(beds_path):
            shutil.rmtree(beds_path)
            generate_beds_by_matrix(cell_feature_adata, beds_path, qpeaks_length_path, n_cores)
        run_info.finish_stage('bed_generation')
    # affinity = get_affinity(cell_feature_adata, total_peaks_path, chip_index, ccre_number)
    ##################################
    ### Search giggle and compute enrich score
    ##################################
    cell_tf_score_df, run_info, _ = search_and_read_result(run_info, beds_path, chip_result_path, chip_index, n_cores)
    ##################################
    ### Summary results
    ##################################
    run_info.finish_stage('result_store')
    if store_result_adata == True:
        cell_tf_score_df.to_csv(result_store_path, sep='\t')
    ##################################
    ### Clean files
    ##################################
    if clean == True:
        try:
            shutil.rmtree(beds_path)
            shutil.rmtree(chip_result_path)
        except:
            pass
    return


def run_enrich(args):
    feature_matrix_path = args.feature_matrix
    species = args.species
    project = args.project
    min_cells = args.min_cells  # for removing few features cells
    min_peaks = args.min_peaks  # for removing few cells features
    max_peaks = args.max_peaks  # for removing doublet cells
    yes = args.yes
    clean = args.clean
    n_cores = args.n_cores

    CONFIG, _ = read_config()
    if species == 'hs':
        chip_index = CONFIG['index']['human_tf_index']
    elif species == 'mm':
        chip_index = CONFIG['index']['mouse_tf_index']
    else:
        pass

    print_log('Reading Files, Please wait ...')
    feature_matrix = sc.read_10x_h5(feature_matrix_path, gex_only=False)

    sc.pp.calculate_qc_metrics(feature_matrix, percent_top=None, log1p=False, inplace=True)
    feature_mean = feature_matrix.obs.n_genes_by_counts.mean()
    feature_std = feature_matrix.obs.n_genes_by_counts.std()

    if min_cells == 'auto':
        min_cells = int(0.005 * feature_matrix.n_obs)
    else:
        min_cells = int(min_cells)
    sc.pp.filter_genes(feature_matrix, min_cells=min_cells)

    if min_peaks == 'auto':
        min_peaks = 500 if feature_mean-3*feature_std < 500 else feature_mean-3*feature_std
    else:
        min_peaks = int(min_peaks)
    sc.pp.filter_cells(feature_matrix, min_genes=min_peaks)  # filter by n_genes_by_counts

    if max_peaks == 'auto':
        max_peaks = feature_mean+3*feature_std
        sc.pp.filter_cells(feature_matrix, max_genes = max_peaks)
        # feature_matrix = feature_matrix[feature_matrix.obs.n_genes_by_counts < max_peaks, :]
    elif max_peaks == 'none':
        print_log('Skip filter by max peaks number.')
    else:
        max_peaks = int(max_peaks)
        sc.pp.filter_cells(feature_matrix, max_genes=max_peaks)
        # feature_matrix = feature_matrix[feature_matrix.obs.n_genes_by_counts < max_peaks, :]


    enrich(feature_matrix,
           species=species,
           project=project,
           chip_index=chip_index,
           store_result_adata=True,
           n_cores=n_cores,
           yes=yes,
           clean=clean
           )

    return



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   setup.py
@Time    :   2021/04/16 12:34:01
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

import os
import sys
import subprocess


def install_giggle():
    prefix = os.environ['PATH'].split(':')[0]
    cmd = f'cd refpkg/giggle; make; cp bin/giggle {prefix}/; cd ../..'
    subprocess.run(cmd, shell=True)

def main():
    install_giggle()
    setup(
        name='SCRIP',
        version='0.0.211219',
        author='Xin Dong',
        author_email='xindong9511@gmail.com',
        description='A package for single cell ATAC-seq analysis',
        packages=['SCRIP', 'SCRIP.enrichment', 'SCRIP.enhancement', 'SCRIP.conf', 'SCRIP.utilities', 'SCRIP.imputation', 'SCRIP.targets'],
        package_data={
            'SCRIP': [
                'conf/config.yml',
                'conf/GRCh38_refgenes.txt',
                'conf/GRCm38_refgenes.txt',
                ],
        },
        install_requires=[
            'numba>=0.51.2',
            'numpy',
            'pandas>=1.1.1',
            'cython>=0.29.22',
            'ruamel.yaml',
            'anndata',
            'anndata2ri',
            # 'Bio',
            'pyranges>=0.0.95',
            'pybedtools>=0.8.2',
            # 'matplotlib',
            # 'seaborn',
            'sklearn',
            'scipy',
            'scanpy>=1.7.1',
        ],
        python_requires='>=3.8.*, !=3.9.*',
        entry_points={
            'console_scripts': [
                'SCRIP=SCRIP.start:main'
            ]
        },
    )



if __name__ == "__main__":
    try:
        from setuptools import setup, find_packages
        main()
    except ImportError:
        print("Can not load setuptools!")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   target.py
@Time    :   2021/05/27 12:46:26
@Author  :   Xin Dong
@Contact :   xindong9511@gmail.com
@License :   (C)Copyright 2020-2021, XinDong
'''

# from multiprocessing import Process, Pool
import os
import scipy
import numpy as np
import pandas as pd
import anndata as ad
import scanpy as sc
from SCRIP.utilities.utils import read_config
# from SCRIP.enrichment.bed_generation import find_nearest_cells

def geneInfoSimple(gene_bed):
    genes_info = []
    genes_list = []
    fhd = open(gene_bed, 'rt')
    fhd.readline()  # skip the first line. In our current gene txt file, there is no '#' in the first line. We need to, perhaps, use the 'ExtractGeneInfo' function.
    for line in fhd:
        line = line.strip().split('\t')
        if not line[0].startswith('#'):
            if line[3] == "+":
                genes_info.append((line[2].replace('chr', ''), int(line[4]), 1, "%s@%s@%s" % (line[12], line[2], line[4])))
            else:
                genes_info.append((line[2].replace('chr', ''), int(line[5]), 1, "%s@%s@%s" % (line[12], line[2], line[5])))
                # gene_info [chrom, tss, 1, gene_unique]
    fhd.close()
    genes_info = list(set(genes_info))
    for igene in range(len(genes_info)):
        tmp_gene = list(genes_info[igene])
        genes_list.append(tmp_gene[3])
        tmp_gene[3] = igene
        genes_info[igene] = tmp_gene
    return genes_info, genes_list

def RP_Simple(peaks_info, genes_info, decay):
    """Multiple processing function to calculate regulation potential."""

    Sg = lambda x: 2**(-x)
    gene_distance = 15 * decay
    genes_peaks_score_array = scipy.sparse.dok_matrix((len(genes_info), len(peaks_info)), dtype=np.float64)

    w = genes_info + peaks_info

    A = {}

    w.sort()
    for elem in w:
        if elem[2] == 1:
            A[elem[-1]] = [elem[0], elem[1]]
        else:
            dlist = []
            for gene_name in list(A.keys()):
                g = A[gene_name]
                tmp_distance = abs(elem[1] - g[1])
                if (g[0] != elem[0]) or (tmp_distance > gene_distance):
                    dlist.append(gene_name)
                else:
                    genes_peaks_score_array[gene_name, elem[-1]] = Sg(tmp_distance / decay)
            for gene_name in dlist:
                del A[gene_name]

    w.reverse()
    for elem in w:
        if elem[2] == 1:
            A[elem[-1]] = [elem[0], elem[1]]
        else:
            dlist = []
            for gene_name in list(A.keys()):
                g = A[gene_name]
                tmp_distance = abs(g[1] - elem[1])
                if (g[0] != elem[0]) or (tmp_distance > gene_distance):
                    dlist.append(gene_name)
                else:
                    genes_peaks_score_array[gene_name, elem[-1]] = Sg(tmp_distance / decay)
            for gene_name in dlist:
                del A[gene_name]

    return genes_peaks_score_array


def count_to_gene_by_RP(input_adata, decay=10000, refgene_path=''):
    cells_list = input_adata.obs.index.tolist()
    peaks_list = input_adata.var.index.tolist()

    genes_info, genes_list = geneInfoSimple(refgene_path)

    peaks_info = []
    for ipeak, peak in enumerate(peaks_list):
        peaks_tmp = peak.rsplit("_", maxsplit=2)
        peaks_info.append([peaks_tmp[0][3:], (int(peaks_tmp[1]) + int(peaks_tmp[2])) / 2.0, 0, ipeak])

    genes_peaks_score_dok = RP_Simple(peaks_info, genes_info, decay)

    genes_peaks_score_csr = genes_peaks_score_dok.tocsr()
    genes_cells_score_csr = genes_peaks_score_csr.dot(input_adata.X.T)

    score_cells_dict = {}
    score_cells_sum_dict = {}

    for igene, gene in enumerate(genes_list):
        score_cells_dict[gene] = igene
        score_cells_sum_dict[gene] = genes_cells_score_csr[igene, :].sum()

    score_cells_dict_dedup = {}
    score_cells_dict_max = {}
    genes = list(set([i.split("@")[0] for i in genes_list]))
    for gene in genes:
        score_cells_dict_max[gene] = float("-inf")

    for gene in genes_list:
        symbol = gene.split("@")[0]
        if score_cells_sum_dict[gene] > score_cells_dict_max[symbol]:
            score_cells_dict_dedup[symbol] = score_cells_dict[gene]
            score_cells_dict_max[symbol] = score_cells_sum_dict[gene]
    gene_symbol = sorted(score_cells_dict_dedup.keys())
    matrix_row = []
    for gene in gene_symbol:
        matrix_row.append(score_cells_dict_dedup[gene])

    score_cells_matrix = genes_cells_score_csr[matrix_row, :]

    RP_adata = ad.AnnData(score_cells_matrix.T, obs=pd.DataFrame(index=cells_list), var=pd.DataFrame(index=gene_symbol))
    return RP_adata


def run_target(args):
    feature_matrix_path = args.feature_matrix
    species = args.species
    output = args.output
    decay = args.decay

    if feature_matrix_path.endswith('.h5'):
        input_mat_adata = sc.read_10x_h5(feature_matrix_path, gex_only=False)
    elif feature_matrix_path.endswith('.h5ad'):
        input_mat_adata = sc.read_h5ad(feature_matrix_path)
    else:
        try:
            input_mat_adata = sc.read_10x_mtx(feature_matrix_path, gex_only=False)
        except KeyError:
            input_mat_adata = sc.read_10x_mtx(feature_matrix_path, var_names='gene_ids' ,gex_only=False)


    CONFIG, CONFIG_PATH = read_config()

    if species == 'hs':
        refgene_path = os.path.join(CONFIG_PATH, 'GRCh38_refgenes.txt')
    elif species == 'mm':
        refgene_path = os.path.join(CONFIG_PATH, 'GRCm38_refgenes.txt')
    else:
        pass
    rp_adata = count_to_gene_by_RP(input_adata=input_mat_adata,
                                   decay=decay,
                                   refgene_path=refgene_path
                                )
    rp_adata.write_h5ad(output)
    return
